#!/usr/bin/env python3

import os
import sys
import pandas as pd
import subprocess
from shutil import copyfile
from collections import Counter
from logzero import logger
from tinydb import TinyDB, Query
from .dictlist import DictList
from .utils import sort_versions


class Tracker(object):
    def __init__(self, path=None):
        """Constructor for class Tracker"""
        self.path = path if path is not None else os.environ['TRACKER_PATH']
        self.db = TinyDB(self.path)
        self.entry = Query()

    def save(self, entry):
        self.db.insert(entry.properties)
        self.deduplicate()
        self.label_recent()

    def filter(self, cond):
        """tr.filter(tr.entry.tag == "tag")"""
        return(self.db.search(cond))

    def uniq(self, property):
        return(set(self[property]))

    def label_recent(self):
        tags = self.uniq('tag')
        for t in tags:
            entries = self.filter(self.entry.tag == t)
            versions = [e['version'] for e in entries]
            most_recent = sort_versions(versions)[-1]
            self.db.update({'most_recent': True}, ((
                self.entry.tag == t) & (self.entry.version == most_recent)))
            self.db.update({'most_recent': False}, ((
                self.entry.tag == t) & (self.entry.version != most_recent)))

    def remove(self, cond):
        """tr.remove(tr.entry.tag == "tag")"""
        logger.info(f"Conditional for removing entries: {cond}")
        self.db.remove(cond)

    def deduplicate(self):
        vtags = [key for key, value in Counter(
            self['tag_version']).items() if value >= 2]
        for tv in vtags:
            entries = self.filter(self.entry.tag_version == tv)
            times = [e['time'] for e in entries]
            most_recent = sorted(times)[-1]
            self.db.remove(((self.entry.tag_version == tv) &
                            (self.entry.time != most_recent)))

    def copy(self, path):
        copyfile(self.path, path)
        return(Tracker(path))

    def get_entry(self, entry_tag, version=None):
        if version is not None:
            entry = self.filter((self.entry.tag == entry_tag) & (
                self.entry.version == version))[0]
        else:
            entry = self.filter((self.entry.tag == entry_tag) & (
                self.entry.most_recent == True))[0]
        return(entry)

    def get_file(self, entry_tag, file_tag, version=None):
        """Get File dictionary given entry tag and file tag

        Parameters
        ----------
        entry_tag : str
            Entry tag in database
        file_tag : str
            File tag identifying specific file
        version : str, optional
            Version string if not going by most recent, by default None

        See Also
        --------
        get_entry, get_file_path

        Examples
        --------
        >>>tr.get_file("entry_tag", "file_tag")
        """
        entry = self.get_entry(entry_tag, version)['output_files']
        dli = DictList(entry)
        return(dli.filter_first(cond=lambda x: x['tag'] == file_tag))

    def get_file_path(self, *args, **kwargs):
        return(self.get_file(*args, **kwargs)['path'])

    def get_output_files(self, entry_tag, version=None):
        """Get list of output files from an entry.

        Parameters
        ----------
        entry_tag : str
            Entry tag in database
        version : str, optional
            Version string if not going by most recent, by default None

        See Also
        --------
        get_entry

        Examples
        --------
        >>>tr.get_output_files("entry_tag")
        """
        return(self.get_entry(entry_tag, version)['output_files'])

    def update(self):
        pass

    def to_pandas(self, tag=None, module=None, most_recent=True):
        df = pd.DataFrame.from_dict([row for row in self.db])
        if df.empty:
            return(df)
        df = df.sort_values('time', ascending=False)
        if tag:
            df = df[df.tag == tag]
        if module:
            df = df[df.module == module]
        if most_recent:
            df = df[df.most_recent]
        return(df)

    def explode(self, *args, **kwargs):
        df = self.to_pandas(*args, **kwargs)

        df = df[['tag', 'category', 'module', 'description',
                 'version', 'input_files', 'output_files', 'most_recent', 'time']]

        df0 = df.drop('output_files', axis=1).rename(
            {'input_files': 'files'}, axis=1)
        df1 = df.drop('input_files', axis=1).rename(
            {'output_files': 'files'}, axis=1)

        df0 = self._explode_files(df0)
        df1 = self._explode_files(df1)

        df = pd.concat([df0.assign(type='input'),
                        df1.assign(type='output')], axis=0)
        df['basename'] = df['path'].apply(os.path.basename)
        df = df[['tag', 'category', 'module', 'file_tag', 'description', 'type',
                 'file_desc', 'basename', 'path', 'most_recent', 'index', 'time']]
        df = df.sort_values(['time', 'category', 'module', 'tag', 'type', 'index'], ascending=[
                            False, True, True, True, True, True])
        return(df)

    def _explode_files(self, df):
        df['files'] = df['files'].map(
            lambda l: [dict(x, **{'index': i}) for i, x in enumerate(l)])
        df = df.explode('files')
        df = df[df.files.notnull()]
        df['file_tag'] = df['files'].map(lambda x: x['tag'])
        df['path'] = df['files'].map(lambda x: x['path'])
        df['file_desc'] = df['files'].map(lambda x: x['description'])
        df['index'] = df['files'].map(lambda x: x['index'])
        df = df.drop('files', axis=1)
        return(df)

    def to_excel(self, path, **kwargs):
        df = self.to_pandas(**kwargs)
        df.to_excel(path)

    def kill(self):
        self.db.truncate()

    def __len__(self):
        return(len(self.db))

    def __getitem__(self, property):
        return([row[property] for row in self.db])

    @property
    def entries(self):
        return(self.db.all())

    @property
    def table(self):
        return(self.to_pandas(tag=None, most_recent=False))

    @property
    def summary(self):
        df = self.table[['category', 'module', 'tag', 'description',
                         'version', 'date', 'time', 'most_recent']]
        return(df.sort_values('time', ascending=False))




#!/usr/bin/env python3

import os
import sys
from logzero import logger

class DictList():
    def __init__(self, dict_list):
        self.dli = dict_list

    def filter(self, key=None, value=None, cond=None):
        if cond is None:
            cond = lambda item: item[key] == value
        return([item for item in self.dli if cond(item)])

    def filter_first(self, key=None, value=None, cond=None):
        if cond is None:
            cond = lambda item: item[key] == value
        return(next((item for item in self.dli if cond(item)), None))

    def filter_first_index(self, key, value, cond=None):
        if cond is None:
            cond = lambda item: item[key] == value
        return(next((i for i, item in enumerate(self.dli) if cond(item)), None))

    def keys(self):
        return(set([key for item in self.dli for key in item]))

    def values(self, key):
        return(set([item[key] for item in self.dli ]))
    
    def __getitem__(self, key):
        if isinstance(key, int):
            return(self.dli[key])
        else:
            return([item[key] for item in self.dli])

    def __len__(self):
        return(len(self.dli))

    def __str__(self):
        return(str(self.dli))

    def __iter__(self):
        return(self.dli)


#!/usr/bin/env python3

import os
import sys
from logzero import logger
import re
import subprocess
import pandas as pd
from pkg_resources import parse_version


def gsutil(*args, silent=True):
    cmd = ['gsutil']
    if len(args) == 1:
        cmd = f'{cmd[0]} {args[0]}'
        shell = True
    else:
        cmd.extend(args)
        shell = False
    p = subprocess.run(cmd, shell=shell, capture_output=True)
    output, error = p.stdout.decode().strip(), p.stderr.decode().strip()
    if not silent:
        log_str = f"{cmd}" if error is b'' else f"{cmd}\n{error}"
        logger.info(log_str)
    return(output)


def gs_ls(path, pattern=None, trimdir=True):
    """gs_ls(tr.get_file_path('rare-gtfilt-genotypes-mt', 'rare-filtered-genotypes-mt'), pattern='.mt$')"""
    output = gsutil(*['ls', path])
    if re.search("\n", output):
        output = output.split("\n")
        if trimdir:
            output = [re.sub('\\/$', '', o) for o in output]
        if pattern:
            output = [o for o in output if re.search(pattern, o)]
        if len(output) == 1:
            output = output[0]
    return(output)


def is_cloud_path(path):
    return(path[0:3] == 'gs:')


def cloud_path_exists(path):
    return(subprocess.run(['gsutil', 'ls', path], capture_output=True).stderr.decode() is '')


def local_path_exists(path):
    return(os.path.exists(os.path.realpath(os.path.expanduser(path))))


def path_exists(path):
    if is_cloud_path(path):
        return(cloud_path_exists(path))
    else:
        return(local_path_exists(path))


def filter_list(arr, pattern):
    arr = [a for a in arr if re.search(pattern, a)]
    if len(arr) == 1:
        return(arr[0])
    else:
        return(arr)


def excel_cache(df, name='summary'):
    """outfile = excel_cache(tr.summary, 'summary'); ! open $outfile"""
    outfile = f'{name}.xlsx'
    df.to_excel(outfile, index=None)
    return(outfile)


def sort_versions(arr):
    """Sort version tags using pkg_resource definition.

    Parameters
    ----------
    arr : list

    Examples
    --------
    >>> versions = ['0.1', '0.10', '0.2.1', '0.2', '0.10.1',
            '0.1.1+0.g3c5592b.dirty', '0.1.5', '0.2', '0.2']
    >>> sort_versions(versions)
    """
    return(sorted(arr, key=parse_version))


#!/usr/bin/env python3

import os
import sys
from logzero import logger
import time
from datetime import date
import subprocess

from .file import InputFile, OutputFile

class Entry():
    def __init__(self, tag, description, category, module, version=None):
        self.properties = {
            'tag': tag,
            'description': description,
            'category': category,
            'module': module,
            'version': self.get_version() if not version else str(version),
            'path': __file__ if 'FILE' not in os.environ else os.environ['FILE'],
            'date': str(date.today()),
            'time': time.time(),
            'input_files': [],
            'output_files': []
        }
        self['tag_version'] = f"{self['tag']}_{self['version']}"
        extension = os.path.splitext(self['path'])[1]

    def get_version(self):
        if 'VERSION' in os.environ:
            version = os.environ['VERSION']
        else:
            version = subprocess.run(['git', 'describe'], capture_output=True).stdout.strip().decode()
        return(version)

    def add(self, file):
        if isinstance(file, OutputFile):
            if file['tag'] == 'Artifact':
                file['tag'] = f"Artifact_{len(self['output_files'])}"
                file['description'] = f"{file['description']} ({len(self['output_files'])})"
            self['output_files'].append(file.properties)
        elif isinstance(file, InputFile):
            self['input_files'].append(file.properties)
        else:
            print('File not in the correct format.')
        return(file['path'])

    def show(self):
        return(self.properties)

    def __getitem__(self, property):
        return(self.properties[property])

    def __setitem__(self, index, value):
        self.properties[index] = value

    def __repr__(self):
        return(str(self.properties))

    def __str__(self):
        return(str(self.properties))

    @property
    def dict(self):
        return(self.properties)


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .tracker import Tracker
from .entry import Entry
from .file import InputFile, OutputFile
from .dictlist import DictList
from .schema import category_template

from .file import path_exists
from .utils import *


#!/usr/bin/env python3

import os
import sys
from datetime import date
import subprocess
from logzero import logger
from .utils import is_cloud_path, path_exists

class File():
    def __init__(self, tag, path, description, source=None):
        self.properties = {
                'tag': tag,
                'path': path,
                'description': description,
                'source': source
        }

    def impute_source(self):
        source = 'cloud' if is_cloud_path(self['path']) else 'local'
        return(source)

    def all(self):
        return(self.properties)

    def __getitem__(self, property):
        return(self.properties[property])
    
    def __setitem__(self, index, value):
        self.properties[index] = value
    
    def __repr__(self):
        return(str(self.properties))
    
    def __str__(self):
        return(str(self['path']))
    
    @property
    def path(self):
        return(self['path'])

class InputFile(File):
    def __init__(self, **kwargs):
        init_dict = { p: kwargs.get(p, None) for p in [ 'path', 'description', 'source' ] }
        init_dict['tag'] = kwargs['tag']
        super().__init__(**init_dict)
        self['entry_tag'] = kwargs.get('entry_tag', None)
        self['version'] = kwargs.get('version', None)
        tracker = kwargs.get('database', None)
        #if tracker is None and 'tr' in globals():
        #    tracker = globals()['tr']
        #print(globals())
        #self.tracker = tracker
        #sum([ self[p] is not None for p in ['entry_tag'] ]) == 3 # , 'version', 'database'
        infer_boolean = (int(tracker is not None) + int(self['entry_tag'] is not None)) == 2
        defined_boolean = sum([ self[p] is not None for p in ['path', 'description'] ]) == 2
        if not infer_boolean and not defined_boolean:
            raise('User must provide either the entry_tag, version, and database reference OR path and description.')

        if infer_boolean:
            self.__infer_file_properties(tracker)

        self['exists'] = path_exists(self['path'])
        
    def __infer_file_properties(self, tracker):
        entry = tracker.get_entry(self['entry_tag'], self['version'])
        file = tracker.get_file(self['entry_tag'], self['tag'], self['version'])
        self['version'] = entry['version']
        self['path'] = file['path']
        self['description'] = file['description']
        self['date'] = file['date']

class OutputFile(File):
    def __init__(self, **kwargs):
        init_dict = { p: kwargs.get(p, None) for p in [ 'tag', 'description', 'source' ] }
        init_dict['path'] = kwargs['path']
        super().__init__(**init_dict)
        self['date'] = str(date.today())
        self['source'] = self.impute_source() if self['source'] is None else self['source']
        if self['tag'] is None:
            self['tag'] = 'Artifact'
            self['description'] = 'Data artifact'


#!/usr/bin/env python3

import os
import sys
from logzero import logger

category_template = {
    'Preparation': [  # 'Pre-Processing'
        'Manifest',
        'Coverage',
        'References',
        'Processing'
    ],
    'Processing': [ # '' Quality Control (QC)
        'Annotation',
        'Variant QC',
        'Sample QC',
        'Calibration'
    ],
    'Post-processing': ['Migration', 'Processing of rare variant counts'],
    'Exploration': [],
    'Analysis': [
        'Gene burden analysis',
        'Gene set enrichment analysis',
        "Permutation-based Fisher's Exact Test",
        "Inherited analysis",
        "De novo mutations",
        "Weighted meta-analysis",
        "Expression",
        "NDDs overlap analysis",
        "Single variant association analysis",
        "Subset and gene plot",
        "Browser",
        "Gene set definition",
        "Genetic architecture of schizophrenia",

    ],
    'Modeling': [],
    'Evaluation': [],
    'Visualization': [],
    'Summary': []
}

entry_properties = {
    'init': [
        # defines entry group
        'tag',
        'description',
        'category',
        'module'
    ],
    'env': [
        'version',
        'script_path',
        'git_hash',
        'date_rendered',
        'core_libraries', # : [ {'library': '', 'version': '' }]
    ],
    'calc': [
        # unique, calculated from tag and version
        'tag_version',
        # from script name
        'language'
    ],
    'update': [
        # calculated, inverse is 'is_retired'
        'is_current',
        # calculated by date and module
        'order'
    ]
}

input_properties = {
    'init': [
        'tag', # user-defined, unique
        'entry_tag',
        'version', # can be default or most recent
        'path', # optional, search in output_properties
        'description', # optional, search in output_properties
        'source', # optional, search in output_properties, cloud or local
    ],
    'calc': [
        'exists'
    ]
}
output_properties = {
    'init': [
        'tag', # user-defined, unique
        'path',
        'description',
        'source', # optional, search in output_properties, cloud or lcal
    ],
    'calc': [
        'date',
        'exists',
        'source'
    ]
}